=begin
Description:

  Provides a Gtk vtkRenderWindowInteractor widget.  This embeds a
  vtkRenderWindow inside a GTK widget && uses the
  vtkGenericRenderWindowInteractor for the event handling.  This is
  vtkGenericRenderWindowInteractor end
  similar to GtkVTKRenderWindowInteractor.rb.

  The extensions here allow the use of gtkglext rather than gtkgl &&
  ruby-gtk2 rather than ruby-gtk.  It requires ruby-gtk2.0.0 || later.

  There is a working example at the bottom.

Credits:

License:

  VTK license.

=end

require 'rbconfig'
require 'gtk2'
require 'gtkglext'
require 'vtk'

class GtkGLExtVTKRenderWindowInteractor < Gtk::DrawingArea
=begin
  Embeds a vtkRenderWindow into a pyGTK widget && uses
  vtkGenericRenderWindowInteractor for the event handling.  This
  class embeds the RenderWindow correctly.  A __getattr__ hook is
  provided that makes the class behave like a
  vtkGenericRenderWindowInteractor.
=end
  def initialize
    super

    glconfig = Gdk::GLConfig.new(Gdk::GLConfig::MODE_RGB|
                                 Gdk::GLConfig::MODE_DEPTH)

    set_gl_capability(glconfig)

    @RenderWindow = Vtk::RenderWindow.new

    # private attributes
    @Created = false
    @ActiveButton = nil

    @Iren = Vtk::GenericRenderWindowInteractor.new
    @Iren.SetRenderWindow(@RenderWindow)
    @Iren.GetInteractorStyle.SetCurrentStyleToTrackballCamera

    createTimer = Proc.new{|obj, event|
      gtk.timeout_add(10, Proc.new{ @Iren.TimerEvent })
    }

    destroyTimer = Proc.new{|obj, event|
=begin
      The timer is a one shot timer so will expire automatically.
=end
      return 1
    }

    @Iren.AddObserver('CreateTimerEvent', createTimer)
    @Iren.AddObserver('DestroyTimerEvent', destroyTimer)
    self.ConnectSignals

    # need this to be able to handle key_press events.
    self.set_flags(Gtk::Window::CAN_FOCUS)
  end

  def set_size_request(w, h)
    super(w,h)
    @RenderWindow.SetSize(w, h)
    @Iren.SetSize(w, h)
    @Iren.ConfigureEvent
  end

  def ConnectSignals
    self.signal_connect("realize"){|wid,event| OnRealize(wid,event) }
    self.signal_connect("expose_event"){|wid,event| OnExpose(wid,event) }
    self.signal_connect("configure_event"){|wid,event| OnConfigure(wid,event) }
    self.signal_connect("button_press_event"){|wid,event| OnButtonDown(wid,event) }
    self.signal_connect("button_release_event"){|wid,event| OnButtonUp(wid,event) }
    self.signal_connect("motion_notify_event"){|wid,event| OnMouseMove(wid,event) }
    self.signal_connect("enter_notify_event"){|wid,event| OnEnter(wid,event) }
    self.signal_connect("leave_notify_event"){|wid,event| OnLeave(wid,event) }
    self.signal_connect("key_press_event"){|wid,event| OnKeyPress(wid,event) }
    self.signal_connect("delete_event"){|wid,event| OnDestroy(wid,event) }
    self.add_events(Gdk::Event::EXPOSURE_MASK |
                    Gdk::Event::BUTTON_PRESS_MASK |
                    Gdk::Event::BUTTON_RELEASE_MASK |
                    Gdk::Event::KEY_PRESS_MASK |
                    Gdk::Event::POINTER_MOTION_MASK |
                    Gdk::Event::POINTER_MOTION_HINT_MASK |
                    Gdk::Event::ENTER_NOTIFY_MASK |
                    Gdk::Event::LEAVE_NOTIFY_MASK)
  end

  def GetRenderWindow
    return @RenderWindow
  end

  def Render
    if @Created
      @RenderWindow.Render
    end
  end

  def OnRealize(wid, event)
    unless @Created
      # you can't get the xid without the window being realized.
      self.realize
      if Config::CONFIG["host_os"] =~ /win32/
        win_id = self.window.handle.to_s
      else
        win_id =self.window.xid.to_s
      end

      @RenderWindow.SetWindowInfo(win_id)
#      @Iren.Initialize
      @Created = true
    end
    return true
  end

  def OnConfigure(widget, event)
#    self.widget=widget
    @Iren.SetSize(event.width, event.height)
    @Iren.ConfigureEvent
    self.Render
    return true
  end

  def OnExpose(wid, event)
    self.Render
    return true
  end

  def OnDestroy(wid, event=nil)
    self.hide
    @RenderWindow = nil
    self.destroy
    return true
  end

  private
  def _GetCtrlShift(event)
    ctrl, shift = 0, 0        
    if ((event.state & Gdk::Window::ModifierType::CONTROL_MASK) == Gdk::Window::ModifierType::CONTROL_MASK)
      ctrl = 1
    end
    if ((event.state & Gdk::Window::ModifierType::SHIFT_MASK) == Gdk::Window::ModifierType::SHIFT_MASK)
      shift = 1
    end
    return ctrl, shift
  end


  public
  def OnButtonDown(wid, event)
=begin
    Mouse button pressed.
=end
    m = self.pointer
    ctrl, shift = _GetCtrlShift(event)
    @Iren.SetEventInformationFlipY(m[0], m[1], ctrl, shift, 0.chr, 0, nil)
    button = event.button
    if button == 3
      @Iren.RightButtonPressEvent
      return true
    elsif button == 1
      @Iren.LeftButtonPressEvent
      return true
    elsif button == 2
      @Iren.MiddleButtonPressEvent
      return true
    else
      return ralse
    end
  end

  def OnButtonUp(wid, event)
=begin
    Mouse button released.
=end
    m = self.pointer
    ctrl, shift = _GetCtrlShift(event)
    @Iren.SetEventInformationFlipY(m[0], m[1], ctrl, shift, 0.chr, 0, nil)
    button = event.button
    if button == 3
      @Iren.RightButtonReleaseEvent
      return true
    elsif button == 1
      @Iren.LeftButtonReleaseEvent
      return true
    elsif button == 2
      @Iren.MiddleButtonReleaseEvent
      return true
    end

    return false
  end

  def OnMouseMove(wid, event)
=begin
    Mouse has moved.
=end
    m = self.pointer
    ctrl, shift = _GetCtrlShift(event)
    @Iren.SetEventInformationFlipY(m[0], m[1], ctrl, shift, 0.chr, 0, nil)
    @Iren.MouseMoveEvent
    return true
  end

  def OnEnter(wid, event)
=begin
    Entering the vtkRenderWindow.
=end
    self.grab_focus
    m = self.pointer
    ctrl, shift = _GetCtrlShift(event)
    @Iren.SetEventInformationFlipY(m[0], m[1], ctrl, shift, 0.chr, 0, nil)
    @Iren.EnterEvent
    return true
  end

  def OnLeave(wid, event)
=begin
    Leaving the vtkRenderWindow.
=end
    m = self.pointer
    ctrl, shift = _GetCtrlShift(event)
    @Iren.SetEventInformationFlipY(m[0], m[1], ctrl, shift, 0.chr, 0, nil)
    @Iren.LeaveEvent
    return true
  end

  def OnKeyPress(wid, event)
=begin
    Key pressed.
=end
    m = self.pointer
    ctrl, shift = _GetCtrlShift(event)
#    keycode, keysym = event.keyval, event.string
    keycode, keysym = event.keyval, nil
    key = 0.chr
    if keycode < 256
      key = keycode.chr
    end
    @Iren.SetEventInformationFlipY(m[0], m[1], ctrl, shift, key, 0, keysym)
    @Iren.KeyPressEvent
    @Iren.CharEvent
    return true
  end

  def OnKeyRelease(wid, event)
=begin
    Key released.
=end
    m = self.pointer
    ctrl, shift = _GetCtrlShift(event)
#    keycode, keysym = event.keyval, event.string
    keycode, keysym = event.keyval, nil
    key = 0.chr
    if keycode < 256
      key = keycode.chr
    end
    @Iren.SetEventInformationFlipY(m[0], m[1], ctrl, shift, key, 0, keysym)
    @Iren.KeyReleaseEvent
    return true
  end

  def Initialize
    if @Created
      @Iren.Initialize
    end
  end

  def Start
    if @Created
      @Iren.Start
    end
  end

  def SetPicker(picker)
    @Iren.SetPicker(picker)
  end

  def GetPicker(picker)
    return @Iren.GetPicker
  end

end

def main

  Gtk.init
  Gtk::GL.init

  # The main window
  window = Gtk::Window.new(Gtk::Window::TOPLEVEL)
  window.set_title("A GtkVTKRenderWindow Demo!")
  window.signal_connect("destroy"){ gtk.main_quit }
  window.signal_connect("delete_event"){ gtk.main_quit }
  window.set_border_width(10)

  # A VBox into which widgets are packed.
  vbox = Gtk::VBox.new(false, 3)
  window.add(vbox)
  vbox.show

  # The GtkVTKRenderWindow
  gvtk = GtkGLExtVTKRenderWindowInteractor.new
  #gvtk.SetDesiredUpdateRate(1000)
  gvtk.set_size_request(400, 400)
  vbox.pack_start(gvtk)
  gvtk.show
  gvtk.Initialize
  gvtk.Start
  # prevents 'q' from exiting the app.
  exit = Proc.new{|obj, event, x|
    x
  }
#  gvtk.AddObserver("ExitEvent", exit)

  # The VTK stuff.
  cone = Vtk::ConeSource.new
  cone.SetResolution(80)
  coneMapper = Vtk::PolyDataMapper.new
  coneMapper.SetInput(cone.GetOutput)
  #coneActor = Vtk::LODActor.new
  coneActor = Vtk::Actor.new
  coneActor.SetMapper(coneMapper)    
  coneActor.GetProperty.SetColor(0.5, 0.5, 1.0)
  ren = Vtk::Renderer.new
  gvtk.GetRenderWindow.AddRenderer(ren)
  ren.AddActor(coneActor)

  # A simple quit button
  quit = Gtk::Button.new("Quit!")
  quit.signal_connect("clicked"){ Gtk.main_quit }
  vbox.pack_start(quit)
  quit.show

    # show the main window && start event processing.
  window.show
  Gtk.main
end


if $0 == __FILE__
  main
end
