=begin
A vtkTkImageViewerWidget for ruby, which is based on the
vtkTkImageWindowWidget.

Specify double=1 to get a double-buffered window.

=end

require 'tk'
require 'vtk'
require 'vtk/tk/vtkLoadRubyTkWidgets'


class VtkTkImageViewerWidget < TkWindow
=begin
    A vtkTkImageViewerWidget for Ruby.

    Use GetImageViewer() to get the vtkImageViewer.

    Create with the keyword double=1 in order to generate a
    double-buffered viewer.
    
    Create with the keyword focus_on_enter=1 to enable
    focus-follows-mouse.  The default is for a click-to-focus mode.    
=end
  TkCommandNames = ['vtkTkImageViewerWidget'.freeze].freeze

  def initialize(master, kw={})
=begin
        Constructor.

        Keyword arguments:

          iv -- Use passed image viewer instead of creating a new one.

          double -- If True, generate a double-buffered viewer.
          Defaults to False.

          focus_on_enter -- If True, use a focus-follows-mouse mode.
          Defaults to False where the widget will use a click-to-focus
          mode.
=end
    # load the necessary extensions into tk
    vtkLoadRubyTkWidgets()

    _symbolkey2str(kw)

    if kw['iv']
      @ImageViewer = kw['iv']
    else
      @ImageViewer = Vtk::ImageViewer.new
    end

    doubleBuffer = 0
    if kw['double']
      doubleBuffer = 1
      kw.delete('double')
    end
 
    # check if focus should follow mouse
    if kw['focus_on_enter']
      @FocusOnEnter = 1
    else
      @FocusOnEnter = 0
    end

    kw['iv'] = @ImageViewer.GetAddressAsString("vtkImageViewer")
    kw['widgetname'] = 'vtkTkImageViewerWidget'
    super(master, kw)

    if doubleBuffer
      @ImageViewer.GetRenderWindow.DoubleBufferOn
    end

    self.BindTkImageViewer
  end

  def GetImageViewer
    return @ImageViewer
  end

  def Render
    @ImageViewer.Render
  end

  def BindTkImageViewer
    imager = @ImageViewer.GetRenderer
        
    # stuff for window level text.
    mapper = Vtk::TextMapper.new
    mapper.SetInput("none")
    t_prop = mapper.GetTextProperty
    t_prop.SetFontFamilyToTimes
    t_prop.SetFontSize(18)
    t_prop.BoldOn
    t_prop.ShadowOn
        
    @LevelMapper = mapper

    actor = Vtk::Actor2D.new
    actor.SetMapper(mapper)
    actor.SetLayerNumber(1)
    actor.GetPositionCoordinate.SetValue(4,22)
    actor.GetProperty.SetColor(1,1,0.5)
    actor.SetVisibility(0)
    imager.AddActor2D(actor)

    @LevelActor = actor
                
    mapper = Vtk::TextMapper.new
    mapper.SetInput("none")
    t_prop = mapper.GetTextProperty
    t_prop.SetFontFamilyToTimes
    t_prop.SetFontSize(18)
    t_prop.BoldOn
    t_prop.ShadowOn
    
    @WindowMapper = mapper

    actor = Vtk::Actor2D.new
    actor.SetMapper(mapper)
    actor.SetLayerNumber(1)
    actor.GetPositionCoordinate.SetValue(4,4)
    actor.GetProperty.SetColor(1,1,0.5)
    actor.SetVisibility(0)
    imager.AddActor2D(actor)

    @WindowActor = actor

    @LastX = 0
    @LastY = 0
    @OldFocus = 0
    @InExpose = 0
    
    # bindings
    # window level
    self.bind("ButtonPress-1"){|e,s| s||s=self; s.StartWindowLevelInteraction(e.x,e.y) }
    self.bind("B1-Motion"){|e,s| s||s=self; s.UpdateWindowLevelInteraction(e.x,e.y) }
    self.bind("ButtonRelease-1"){|e,s| s||s=self; s.EndWindowLevelInteraction }
    
    # Get the value
    self.bind("ButtonPress-3"){|e,s| s||s=self; s.StartQueryInteraction(e.x,e.y) }
    self.bind("B3-Motion"){|e,s| s||s=self; s.UpdateQueryInteraction(e.x,e.y) }
    self.bind("ButtonRelease-3"){|e,s| s||s=self; s.EndQueryInteraction }

    self.bind("Expose"){|e,s| s||s=self; s.ExposeTkImageViewer }
    self.bind("Enter"){|e,s| s||s=self; s.EnterTkViewer }
    self.bind("Leave"){|e,s| s||s=self; s.LeaveTkViewer }
    self.bind("KeyPress-e"){|e,s| s||s=self; exit }
    self.bind("KeyPress-r"){|e,s| s||s=self; s.ResetTkImageViewer }
  end

  def GetImageViewer
    return @ImageViewer
  end

  def Render
    return @ImageViewer.Render
  end

  def _GrabFocus
    @OldFocus = self.focus
    self.focus
  end
        
  def EnterTkViewer
    if @FocusOnEnter
      self._GrabFocus
    end
    return nil
  end

  def LeaveTkViewer
    if @FocusOnEnter && @OldFocus != nil
      @OldFocus.focus
    end
    return nil
  end

  def ExposeTkImageViewer
    if @InExpose == 0
      @InExpose = 1
      self.update()
      @ImageViewer.Render()
      @InExpose = 0
    end
    return nil
  end

  def StartWindowLevelInteraction(x,y)
    if ! @FocusOnEnter
      _GrabFocus
    end
    viewer = @ImageViewer
    @LastX = x
    @LastY = y
    @Window = viewer.GetColorWindow.to_f
    @Level = viewer.GetColorLevel.to_f

    # make the window level text visible
    @LevelActor.SetVisibility(1)
    @WindowActor.SetVisibility(1)

    self.UpdateWindowLevelInteraction(x,y)
    return nil
  end

  def EndWindowLevelInteraction
    # make the window level text invisible
    @LevelActor.SetVisibility(0)
    @WindowActor.SetVisibility(0)
    self.Render
  end

  def UpdateWindowLevelInteraction(x,y)
    # compute normalized delta
    dx = 4.0*(x - @LastX)/self.winfo_width*@Window
    dy = 4.0*(@LastY - y)/self.winfo_height*@Level

    # abs so that direction does not flip
    if @Window < 0.0
      dx = -dx
    end
    if @Level < 0.0
      dy = -dy
    end

    # compute new window level
    window = @Window + dx
    if window < 0.0
      level = @Level + dy
    else
      level = @Level - dy
    end

    viewer = @ImageViewer
    viewer.SetColorWindow(window)
    viewer.SetColorLevel(level)

    @WindowMapper.SetInput("Window: %g" % window)
    @LevelMapper.SetInput("Level: %g" % level)
    
    self.Render
  end

  def ResetTkImageViewer
    # Reset: Set window level to show all values
    viewer = @ImageViewer
    input = viewer.GetInput
    if input == nil
      return nil
    end

    # Get the extent in viewer
    z = viewer.GetZSlice

    input.SetUpdateExtent(-99999,99999,-99999,99999,z,z)
    input.Update

    low,high = input.GetScalarRange
   
    viewer.SetColorWindow(high - low)
    viewer.SetColorLevel((high + low) * 0.5)

    self.Render
  end

  def StartQueryInteraction(x,y)
    if ! @FocusOnEnter
      _GrabFocus()
    end
    # Query PixleValue stuff
    @WindowActor.SetVisibility(1)
    self.UpdateQueryInteraction(x,y)
  end

  def EndQueryInteraction
    @WindowActor.SetVisibility(0)
    self.Render
  end

  def UpdateQueryInteraction(x,y)
    viewer = @ImageViewer
    input = viewer.GetInput
    z = viewer.GetZSlice

    # y is flipped upside down
    y = self.winfo_height - y

    # make sure point is in the whole extent of the image.
    xMin,xMax,yMin,yMax,zMin,zMax = input.GetWholeExtent
    if (x < xMin || x > xMax || y < yMin || y > yMax || z < zMin || z > zMax)
      return nil
    end

    input.SetUpdateExtent(x,x,y,y,z,z)
    input.Update
    numComps = input.GetNumberOfScalarComponents
    text = ""
    for i in 0...numComps
      val = input.GetScalarComponentAsDouble(x,y,z,i)
      text = "#{text}  #{"%.1f"%val}"
    end

    @WindowMapper.SetInput("(#{"%d"%x}, #{"%d"%y}): #{text}")
        
    self.Render
  end

end

#-----------------------------------------------------------------------------
# an example of how to use this widget
if $0 == __FILE__
  canvas = Vtk::ImageCanvasSource2D.new
  canvas.SetNumberOfScalarComponents(3)
  canvas.SetScalarType(3)
  canvas.SetExtent(0,511,0,511,0,0)
  canvas.SetDrawColor(100,100,0)
  canvas.FillBox(0,511,0,511)
  canvas.SetDrawColor(200,0,200)
  canvas.FillBox(32,511,100,500)
  canvas.SetDrawColor(100,0,0)
  canvas.FillTube(550,20,30,400,5)
  canvas.SetDrawColor(255,255,255)
  canvas.DrawSegment3D(10,20,0,90,510,0)
  canvas.SetDrawColor(200,50,50)
  canvas.DrawSegment3D(510,90,0,10,20,0)

  # Check segment clipping
  canvas.SetDrawColor(0,200,0)
  canvas.DrawSegment(-10,30,30,-10)
  canvas.DrawSegment(-10,481,30,521)
  canvas.DrawSegment(481,-10,521,30)
  canvas.DrawSegment(481,521,521,481)

  # Check Filling a triangle
  canvas.SetDrawColor(20,200,200)
  canvas.FillTriangle(-100,100,190,150,40,300)

  # Check drawing a circle
  canvas.SetDrawColor(250,250,10)
  canvas.DrawCircle(350,350,200.0)

  # Check drawing a point
  canvas.SetDrawColor(250,250,250)
  canvas.DrawPoint(350,350)
  canvas.DrawPoint(350,550)

  # Test filling functionality
  canvas.SetDrawColor(55,0,0)
  canvas.DrawCircle(450,350,80.0)
  canvas.SetDrawColor(100,255,100)
  canvas.FillPixel(450,350)

  # Create the GUI: two renderer widgets and a quit button

  frame = TkFrame.new

  widget = VtkTkImageViewerWidget.new(frame,'width'=>512,'height'=>512,'double'=>1)
  viewer = widget.GetImageViewer
  viewer.SetInput(canvas.GetOutput)
  viewer.SetColorWindow(256)
  viewer.SetColorLevel(127.5)

  button = TkButton.new(frame){
    text "Quit"
    command { exit }
  }

  widget.pack('side'=>'top','padx'=>3,'pady'=>3,'fill'=>'both','expand'=>true)
  frame.pack('fill'=>'both','expand'=>true)
  button.pack('fill'=>'x')

  frame.mainloop

end
